@@ -1,4 +1,4 @@
-# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -26,6 +26,7 @@ from __future__ import print_function
 import codecs
 import re
 import os
+import time
 
 import renpy.display
 import renpy.test
@@ -152,6 +153,7 @@ def elide_filename(fn):
     or relative to the Ren'Py directory.
     """
 
+    ofn = fn
     fn = os.path.abspath(fn)
     basedir = os.path.abspath(renpy.config.basedir)
     renpy_base = os.path.abspath(renpy.config.renpy_base)
@@ -161,7 +163,7 @@ def elide_filename(fn):
     elif fn.startswith(renpy_base):
         return os.path.relpath(fn, renpy_base).replace("\\", "/")
     else:
-        return fn.replace("\\", "/")
+        return ofn.replace("\\", "/")
 
 
 def unelide_filename(fn):
@@ -169,7 +171,7 @@ def unelide_filename(fn):
     if os.path.exists(fn1):
         return fn1
 
-    fn2 = os.path.join(renpy.config.basedir, fn)
+    fn2 = os.path.join(renpy.config.renpy_base, fn)
     if os.path.exists(fn2):
         return fn2
 
@@ -190,6 +192,17 @@ def list_logical_lines(filename, filedata=None, linenumber=1, add_lines=False):
     contents. In that case, `filename` need not exist.
     """
 
+    def munge_string(m):
+        brackets = m.group(1)
+
+        if (len(brackets) & 1) == 0:
+            return m.group(0)
+
+        if "__" in m.group(2):
+            return m.group(0)
+
+        return brackets + prefix + m.group(2)
+
     global original_filename
 
     original_filename = filename
@@ -331,6 +344,8 @@ def list_logical_lines(filename, filedata=None, linenumber=1, add_lines=False):
                     pos += 2
                     triplequote = True
 
+                s = [ ]
+
                 while pos < len_data:
 
                     c = data[pos]
@@ -345,31 +360,40 @@ def list_logical_lines(filename, filedata=None, linenumber=1, add_lines=False):
                     if escape:
                         escape = False
                         pos += 1
-                        line.append(c)
+                        s.append(c)
                         continue
 
                     if c == delim:
 
                         if not triplequote:
                             pos += 1
-                            line.append(c)
+                            s.append(c)
                             break
 
                         if (pos < len_data - 2) and (data[pos+1] == delim) and (data[pos+2] == delim):
                             pos += 3
-                            line.append(delim)
-                            line.append(delim)
-                            line.append(delim)
+                            s.append(delim)
+                            s.append(delim)
+                            s.append(delim)
                             break
 
                     if c == u'\\':
                         escape = True
 
-                    line.append(c)
+                    s.append(c)
                     pos += 1
 
                     continue
 
+                s = "".join(s)
+
+                if "[__" in s:
+
+                    # Munge substitutions.
+                    s = re.sub(r'(\.|\[+)__(\w+)', munge_string, s)
+
+                line.append(s)
+
                 continue
 
             word, magic, end = match_logical_word(data, pos)
@@ -534,7 +558,7 @@ class Lexer(object):
     sub-lexers to lex sub-blocks.
     """
 
-    def __init__(self, block, init=False, init_offset=0, global_label=None):
+    def __init__(self, block, init=False, init_offset=0, global_label=None, monologue_delimiter="\n\n"):
 
         # Are we underneath an init block?
         self.init = init
@@ -558,6 +582,8 @@ class Lexer(object):
         self.word_cache_newpos = -1
         self.word_cache = ""
 
+        self.monologue_delimiter = monologue_delimiter
+
     def advance(self):
         """
         Advances this lexer to the next line in the block. The lexer
@@ -692,7 +718,7 @@ class Lexer(object):
 
         init = self.init or init
 
-        return Lexer(self.subblock, init=init, init_offset=self.init_offset, global_label=self.global_label)
+        return Lexer(self.subblock, init=init, init_offset=self.init_offset, global_label=self.global_label, monologue_delimiter=self.monologue_delimiter)
 
     def string(self):
         """
@@ -751,6 +777,75 @@ class Lexer(object):
 
         return s
 
+    def triple_string(self):
+        """
+        Lexes a triple quoted string, intended for use with monologue mode.
+        This is about the same as the double-quoted strings, except that
+        runs of whitespace with multiple newlines are turned into a single
+        newline.
+        """
+
+        s = self.match(r'r?"""([^\\"]|\\.)*"""')
+
+        if s is None:
+            s = self.match(r"r?'''([^\\']|\\.)*'''")
+
+        if s is None:
+            s = self.match(r"r?```([^\\`]|\\.)*```")
+
+        if s is None:
+            return None
+
+        if s[0] == 'r':
+            raw = True
+            s = s[1:]
+        else:
+            raw = False
+
+        # Strip off delimiters.
+        s = s[3:-3]
+
+        def dequote(m):
+            c = m.group(1)
+
+            if c == "{":
+                return "{{"
+            elif c == "[":
+                return "[["
+            elif c == "%":
+                return "%%"
+            elif c == "n":
+                return "\n"
+            elif c[0] == 'u':
+                group2 = m.group(2)
+
+                if group2:
+                    return unichr(int(m.group(2), 16))
+            else:
+                return c
+
+        if not raw:
+
+            # Collapse runs of whitespace into single spaces.
+            s = re.sub(r' *\n *', '\n', s)
+
+            rv = [ ]
+
+            for s in s.split(self.monologue_delimiter):
+                s = s.strip()
+
+                if not s:
+                    continue
+
+                s = re.sub(r'\s+', ' ', s)
+                s = re.sub(r'\\(u([0-9a-fA-F]{1,4})|.)', dequote, s)
+
+                rv.append(s)
+
+            return rv
+
+        return s
+
     def integer(self):
         """
         Tries to parse an integer. Returns a string containing the
@@ -798,6 +893,11 @@ class Lexer(object):
         oldpos = self.pos
         rv = self.word()
 
+        if (rv == "r") or (rv == "u"):
+            if self.text[self.pos:self.pos+1] in ( '"', "'", "`"):
+                self.pos = oldpos
+                return None
+
         if rv in KEYWORDS:
             self.pos = oldpos
             return None
@@ -868,6 +968,11 @@ class Lexer(object):
         oldpos = self.pos
         rv = self.match(image_word_regexp)
 
+        if (rv == "r") or (rv == "u"):
+            if self.text[self.pos:self.pos+1] in ( '"', "'", "`"):
+                self.pos = oldpos
+                return None
+
         if rv in KEYWORDS:
             self.pos = oldpos
             return None
@@ -948,7 +1053,13 @@ class Lexer(object):
 
         return rv
 
-    def delimited_python(self, delim):
+    def expr(self, s, expr):
+        if not expr:
+            return s
+
+        return renpy.ast.PyExpr(s, self.filename, self.number)
+
+    def delimited_python(self, delim, expr=True):
         """
         This matches python code up to, but not including, the non-whitespace
         delimiter characters. Returns a string containing the matched code,
@@ -963,7 +1074,7 @@ class Lexer(object):
             c = self.text[self.pos]
 
             if c in delim:
-                return renpy.ast.PyExpr(self.text[start:self.pos], self.filename, self.number)
+                return self.expr(self.text[start:self.pos], expr)
 
             if c in "'\"":
                 self.python_string()
@@ -976,18 +1087,18 @@ class Lexer(object):
 
         self.error("reached end of line when expecting '%s'." % delim)
 
-    def python_expression(self):
+    def python_expression(self, expr=True):
         """
         Returns a python expression, which is arbitrary python code
         extending to a colon.
         """
 
-        pe = self.delimited_python(':')
+        pe = self.delimited_python(':', False)
 
         if not pe:
             self.error("expected python_expression")
 
-        rv = renpy.ast.PyExpr(pe.strip(), pe.filename, pe.linenumber)  # E1101
+        rv = self.expr(pe.strip(), expr)  # E1101
 
         return rv
 
@@ -1002,19 +1113,19 @@ class Lexer(object):
 
         if c == '(':
             self.pos += 1
-            self.delimited_python(')')
+            self.delimited_python(')', False)
             self.pos += 1
             return True
 
         if c == '[':
             self.pos += 1
-            self.delimited_python(']')
+            self.delimited_python(']', False)
             self.pos += 1
             return True
 
         if c == '{':
             self.pos += 1
-            self.delimited_python('}')
+            self.delimited_python('}', False)
             self.pos += 1
             return True
 
@@ -1154,6 +1265,15 @@ class Lexer(object):
         self.pos = len(self.text)
         return renpy.ast.PyExpr(self.text[pos:].strip(), self.filename, self.number)
 
+    def rest_statement(self):
+        """
+        Like rest, but returns a string rather than a PyExpr.
+        """
+
+        pos = self.pos
+        self.pos = len(self.text)
+        return self.text[pos:].strip()
+
     def python_block(self):
         """
         Returns the subblock of this code, and subblocks of that
@@ -2019,7 +2139,7 @@ def transform_statement(l, loc):
 
 @statement("$")
 def one_line_python(l, loc):
-    python_code = l.rest()
+    python_code = l.rest_statement()
 
     if not python_code:
         l.error('expected python code')
@@ -2133,6 +2253,23 @@ def init_statement(l, loc):
     return ast.Init(loc, block, priority + l.init_offset)
 
 
+@statement("rpy monologue")
+def rpy_statement(l, loc):
+
+    if l.keyword("double"):
+        l.monologue_delimiter = "\n\n"
+    elif l.keyword("single"):
+        l.monologue_delimiter = "\n"
+    else:
+        l.error("rpy monologue expects either single or double.")
+
+    l.expect_eol()
+    l.expect_noblock('rpy monologue')
+    l.advance()
+
+    return [ ]
+
+
 def screen1_statement(l, loc):
 
     # The guts of screen language parsing is in screenlang.py. It
@@ -2238,7 +2375,11 @@ def translate_strings(init_loc, language, l):
                 ll.error("previous string is missing a translation")
 
             loc = ll.get_location()
-            old = parse_string(ll.rest())
+
+            try:
+                old = parse_string(ll.rest())
+            except:
+                ll.error("Could not parse string.")
 
         elif ll.keyword('new'):
 
@@ -2246,7 +2387,10 @@ def translate_strings(init_loc, language, l):
                 ll.error('no string to translate')
 
             newloc = ll.get_location()
-            new = parse_string(ll.rest())
+            try:
+                new = parse_string(ll.rest())
+            except:
+                ll.error("Could not parse string.")
 
             block.append(renpy.ast.TranslateString(loc, language, old, new, newloc))
 
@@ -2435,7 +2579,21 @@ def finish_say(l, loc, who, what, attributes=None):
 
             arguments = args
 
-    return ast.Say(loc, who, what, with_, attributes=attributes, interact=interact, arguments=arguments)
+    if isinstance(what, list):
+
+        rv = [ ]
+
+        for i in what:
+
+            if i == "{clear}":
+                rv.append(ast.UserStatement(loc, "nvl clear", [ ]))
+            else:
+                rv.append(ast.Say(loc, who, i, with_, attributes=attributes, interact=interact, arguments=arguments))
+
+        return rv
+
+    else:
+        return ast.Say(loc, who, what, with_, attributes=attributes, interact=interact, arguments=arguments)
 
 
 @statement("")
@@ -2444,7 +2602,7 @@ def say_statement(l, loc):
     state = l.checkpoint()
 
     # Try for a single-argument say statement.
-    what = l.string()
+    what = l.triple_string() or l.string()
 
     rv = finish_say(l, loc, None, what)
 
@@ -2479,7 +2637,7 @@ def say_statement(l, loc):
     else:
         attributes = None
 
-    what = l.string()
+    what = l.triple_string() or l.string()
 
     if (who is not None) and (what is not None):
 
@@ -2618,6 +2776,7 @@ def report_parse_errors():
 
     print(file=f)
     print("Ren'Py Version:", renpy.version, file=f)
+    print(time.ctime(), file=f)
 
     f.close()
 
